#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from sqlalchemy_tools.utils.text import camel_case_to_underscore

from entity_read.entity import Entity


def ensure_tablename(entity_type: type[Entity], table_name: str = None) -> None:
    if table_name:
        entity_type.__tablename__ = table_name
    if not entity_type.__tablename__:
        entity_type.__tablename__ = camel_case_to_underscore(entity_type.__name__)


def set_table_schema(entity_type: type[Entity], table_schema: str = None) -> None:
    if table_schema is None:
        return
    entity_type.__table_schema_name__ = table_schema
    table_args = getattr(entity_type, '__table_args__', None)
    if isinstance(table_args, tuple):
        if isinstance(table_args[-1], dict):
            target_dict = table_args[-1]
        else:
            entity_type.__table_args__ = table_args + (target_dict := {}, )
    elif isinstance(table_args, dict):
        target_dict = table_args
    elif table_args is None:
        entity_type.__table_args__ = (target_dict := {})
    else:
        raise TypeError(f'Table has "__table_args__" with unexpected type {type(table_args)}')

    if schema := target_dict.get('schema'):
        raise KeyError(f'Attempt to specify schema "{table_schema}", but entity "{entity_type.__name__}" '
                       f'already has schema specified as "{schema}"')
    target_dict["schema"] = table_schema
    print()



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses


@dataclasses.dataclass(frozen=True)
class _NoDefault:
    """
    It should be equal `dataclasses.MISSING` to be ignored by dict_to_dataclass
    But shouldn't be `dataclasses.MISSING`, cos sqlalchemy breaks =(
    """
    def __eq__(self, other):
        return dataclasses.MISSING == other

    def __repr__(self):
        return f'{type(self).__name__}()'

    def __call__(self, *args, **kwargs):
        return self

    def __bool__(self):
        return False


NoDefault = _NoDefault()



from dataclasses import dataclass, field
from typing import ClassVar, Any

from entity_read.sql.atoms import Function, Selectable


@dataclass(kw_only=True, frozen=True)
class FunctionJsonGet(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'json_get'

    def _to_sql(self, args_sql: list) -> str:
        return f"{args_sql[0]} ->> {args_sql[1]}"


@dataclass(kw_only=True, frozen=True)
class FunctionJsonHas(Function):
    type_: str = field(default=str, init=False)
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'json_has'

    def _to_sql(self, args_sql: list):
        return f"{args_sql[0]} ? {args_sql[1]}"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .bool_func import (
    FunctionOr, FunctionAnd, FunctionNot, FunctionIn, FunctionNotIn,
    FunctionLessThan, FunctionLessEqual, FunctionEqual, FunctionNonEqual, FunctionGreaterEqual, FunctionGreater,
    FunctionIsNull, FunctionIsNotNull, FunctionStringContains, FunctionBetween
)
from .other_func import (
    FunctionGetSubqueryLength, FunctionCastToStr, FunctionExists, FunctionNotExists, FunctionDiv, FunctionSub
)
from .agg_func import AggregationArray, AggregationCount
from .json_func import FunctionJsonGet, FunctionJsonHas



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import Field
from typing import TypeVar

from dynamic_types.class_name import _prepare_class_name

from .abstract import AbstractView
from ..entity import Entity

T = TypeVar("T")


class EntityView(AbstractView, abc.ABC):
    @classmethod
    def _produce_view(cls, entity_type: type[Entity]) -> type:
        if not isinstance(entity_type, type) or not issubclass(entity_type, Entity):
            raise TypeError(f'EntityView subscription requires type[Entity], got: {entity_type}')
        field_name_to_field = cls._produce_field_name_to_field(entity_type)
        class_name = _prepare_class_name([entity_type], cls)
        return cls._produce_child_dataclass(class_name, field_name_to_field)

    @classmethod
    @abc.abstractmethod
    def _produce_field_name_to_field(cls, entity_type: type[Entity]) -> dict[str, Field]:
        pass



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import itertools
from dataclasses import dataclass, field
from typing import ClassVar

from entity_read.sql.atoms.node import Node
from entity_read.sql.context import SqlContext
from .abstract import Selectable
from .column import Column


@dataclass(kw_only=True, frozen=True, slots=True)
class Function(Selectable, abc.ABC):
    args: tuple[Selectable, ...]
    code: ClassVar[str]
    keep_subquery: ClassVar[bool] = False

    def requires(self) -> list[Column]:
        return list(itertools.chain(*[arg.requires() for arg in self.args if isinstance(arg, Node)]))
    
    def get_name(self) -> str:
        return f'{self.code}({",".join([arg.get_name() for arg in self.args])})'

    def to_sql(self, context: SqlContext, key_to_expression: dict = None) -> str:
        return "(" + self._to_sql([arg.to_sql(context, key_to_expression) for arg in self.args]) + ")"

    @abc.abstractmethod
    def _to_sql(self, args_sql: list[str]) -> str:
        ...
